# Copyright 2011 - 2019 Patrick Ulbrich <zulu99@gmx.net>
# Copyright 2016 Timo Kankare <timo.kankare@iki.fi>
# Copyright 2016 Thomas Haider <t.haider@deprecate.de>
# Copyright 2011 Ralf Hersel <ralf.hersel@gmx.net>
# Copyright 2019 razer <razerraz@free.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#

"""Implementation for POP3 mailbox connection."""

import email
import logging
import poplib

from Mailnag.backends.base import MailboxBackend
from Mailnag.common.exceptions import InvalidOperationException


class POP3MailboxBackend(MailboxBackend):
	"""Implementation of POP3 mail boxes."""
	
	def __init__(self, name = '', user = '', password = '', oauth2string = '',
				 server = '', port = '', ssl = True, **kw):
		self.name = name
		self.user = user
		self.password = password
		self.oauth2string = oauth2string
		self.server = server
		self.port = port
		self.ssl = ssl # bool
		self._conn = None


	def open(self):
		if self._conn != None:
			raise InvalidOperationException("Account is aready open")
				
		conn = None
		
		try:
			if self.ssl:
				if self.port == '':
					conn = poplib.POP3_SSL(self.server)
				else:
					conn = poplib.POP3_SSL(self.server, int(self.port))
			else:
				if self.port == '':
					conn = poplib.POP3(self.server)
				else:
					conn = poplib.POP3(self.server, int(self.port))
				
				try:
					conn.stls()
				except:
					logging.warning("Using unencrypted connection for account '%s'" % self.name)
				
			conn.getwelcome()
			conn.user(self.user)
			conn.pass_(self.password)
		except:
			try:
				if conn != None:
					conn.quit()
			except:	pass
			raise # re-throw exception
		
		self._conn = conn


	def close(self):
		if self._conn != None:
			self._conn.quit()
			self._conn = None


	def is_open(self):
		return (self._conn != None) and \
				('sock' in self._conn.__dict__)


	def list_messages(self):
		self._ensure_open()
		
		conn = self._conn
		folder = ''
		
		# number of mails on the server
		mail_total = len(conn.list()[1])
		for i in range(1, mail_total + 1): # for each mail
			try:
				# header plus first 0 lines from body
				message = conn.top(i, 0)[1]
			except:
				logging.debug("Couldn't get POP message.")
				continue
			
			# convert list to byte sequence
			message_bytes = b'\n'.join(message)
			
			try:
				# put message into email object and make a dictionary
				msg = dict(email.message_from_bytes(message_bytes))
			except:
				logging.debug("Couldn't get msg from POP message.")
				continue
			yield (folder, msg, {})


	def request_folders(self):
		raise NotImplementedError("POP3 does not support folders")


	def supports_mark_as_seen(self):
		return False


	def mark_as_seen(self, mails):
		raise NotImplementedError


	def notify_next_change(self, callback=None, timeout=None):
		raise NotImplementedError("POP3 does not support notifications")


	def cancel_notifications(self):
		raise NotImplementedError("POP3 does not support notifications")


	def _ensure_open(self):
		if not self.is_open():
			raise InvalidOperationException("Account is not open")



# Copyright 2016 Timo Kankare <timo.kankare@iki.fi>
# Copyright 2014 - 2021 Patrick Ulbrich <zulu99@gmx.net>
# Copyright 2020 Andreas Angerer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#

import threading
import logging
import time

from Mailnag.common.accounts import AccountManager
from Mailnag.daemon.mailchecker import MailChecker
from Mailnag.daemon.mails import Memorizer
from Mailnag.daemon.idlers import IdlerRunner
from Mailnag.daemon.conntest import ConnectivityTest, TestModes
from Mailnag.daemon.dbus import DBusService
from Mailnag.common.plugins import Plugin, HookRegistry, HookTypes, MailnagController
from Mailnag.common.exceptions import InvalidOperationException
from Mailnag.common.config import read_cfg
from Mailnag.common.utils import try_call


testmode_mapping = {
	'auto'				: TestModes.NETWORKMONITOR,	# Legacy, deprecated
	'networkmanager'	: TestModes.NETWORKMONITOR, # Legacy, deprecated
	'networkmonitor'	: TestModes.NETWORKMONITOR,
	'ping'				: TestModes.PING
}

class MailnagDaemon(MailnagController):
	def __init__(self, fatal_error_handler = None, shutdown_request_handler = None):
		self._fatal_error_handler = fatal_error_handler
		self._shutdown_request_handler = shutdown_request_handler
		self._plugins = []
		self._poll_thread = None
		self._poll_thread_stop = threading.Event()
		self._idlrunner = None
		self._disposed = False
	
		self._cfg = read_cfg()
		
		accountman = AccountManager()
		accountman.load_from_cfg(self._cfg, enabled_only = True)
		self._accounts = accountman.to_list()
		self._hookreg = HookRegistry()
		self._conntest = ConnectivityTest(testmode_mapping[self._cfg.get('core', 'connectivity_test')])
		
		self._memorizer = Memorizer()
		self._memorizer.load()
		
		dbus_service = DBusService(self)

		self._mailchecker = MailChecker(self._cfg, self._memorizer, self._hookreg, self._conntest, dbus_service)

		# Note: all code following _load_plugins() should be executed
		# asynchronously because the dbus plugin requires an active mainloop
		# (usually started in the programs main function).
		self._load_plugins(self._cfg)

		# Start checking for mails asynchronously.
		self._start_thread = threading.Thread(target = self._start)
		self._start_thread.start()
	
	
	def dispose(self):
		if self._disposed:
			return

		# Note: _disposed must be set 
		# before cleaning up resources 
		# (in case an exception occurs) 
		# and before unloading plugins.
		# Also required by _wait_for_inet_connection().
		self._disposed = True
		
		# clean up resources
		if (self._start_thread != None) and (self._start_thread.is_alive()):
			self._start_thread.join()
			logging.info('Starter thread exited successfully.')

		if (self._poll_thread != None) and (self._poll_thread.is_alive()):
			self._poll_thread_stop.set()
			self._poll_thread.join()
			logging.info('Polling thread exited successfully.')

		if self._idlrunner != None:
			self._idlrunner.dispose()

		if self._accounts != None:
			for acc in self._accounts:
				if acc.is_open():
					try: acc.close()
					except: pass

		self._unload_plugins()
	
	
	def is_disposed(self):
		return self._disposed
	
	
	# Part of MailnagController interface
	def get_hooks(self):
		# Note: ensure_not_disposed() cannot be called here
		# because plugins are calling get_hooks() when being disabled in dispose().
		# self._ensure_not_disposed()
		return self._hookreg
	

	# Part of MailnagController interface
	def shutdown(self):
		if self._shutdown_request_handler != None:
			self._shutdown_request_handler()
	
	
	# Part of MailnagController interface
	def mark_mail_as_read(self, mail_id):
		# Note: ensure_not_disposed() is not really necessary here
		# (the memorizer object is available in dispose()), 
		# but better be consistent with other daemon methods.
		self._ensure_not_disposed()
		
		self._memorizer.set_to_seen(mail_id)
		self._memorizer.save()
	
	
	# Enforces manual mail checks
	# Part of MailnagController interface
	def check_for_mails(self):
		# Don't allow mail checks after object disposal. 
		# F.i. plugins may not be unloaded completely or 
		# connections may have been closed already.
		self._ensure_not_disposed()
		
		non_idle_accounts = self._get_non_idle_accounts(self._accounts)
		self._mailchecker.check(non_idle_accounts)

		
	def _ensure_not_disposed(self):
		if self._disposed:
			raise InvalidOperationException(
				"Daemon has been disposed")
	
	
	def _start(self):
		try:
			# Call Accounts-Loaded plugin hooks
			for f in self._hookreg.get_hook_funcs(HookTypes.ACCOUNTS_LOADED):
				try_call( lambda: f(self._accounts) )
			
			if not self._wait_for_inet_connection():
				return
			
			# Immediate check, check *all* accounts
			try:
				self._mailchecker.check(self._accounts)
			except:
				logging.exception('Caught an exception.')

			idle_accounts = self._get_idle_accounts(self._accounts)
			non_idle_accounts = self._get_non_idle_accounts(self._accounts)

			# start polling thread for POP3 accounts and
			# IMAP accounts without idle support
			if len(non_idle_accounts) > 0:
				poll_interval = int(self._cfg.get('core', 'poll_interval'))

				def poll_func():
					while True:
						self._poll_thread_stop.wait(timeout = 60.0 * poll_interval)
						if self._poll_thread_stop.is_set():
							break
						
						try:
							self._mailchecker.check(non_idle_accounts)
						except:
							logging.exception('Caught an exception.')

				self._poll_thread = threading.Thread(target = poll_func)
				self._poll_thread.start()

			# start idler threads for IMAP accounts with idle support
			if len(idle_accounts) > 0:
				def sync_func(account):
					self._mailchecker.check([account])


				idle_timeout = int(self._cfg.get('core', 'imap_idle_timeout'))				
				self._idlrunner = IdlerRunner(idle_accounts, sync_func, idle_timeout)
				self._idlrunner.start()
		except Exception as ex:
			logging.exception('Caught an exception.')
			if self._fatal_error_handler != None:
				self._fatal_error_handler(ex)
	

	def _wait_for_inet_connection(self):
		if self._conntest.is_offline():
			logging.info('Waiting for internet connection...')
		
		while True:
			if self._disposed:					return False
			if not self._conntest.is_offline():	return True
			# Note: don't sleep too long
			# (see timeout in mailnag.cleanup())
			# ..but also don't sleep to short in case of a ping connection test.
			time.sleep(3)


	def _get_idle_accounts(self, accounts):
		return [acc for acc in self._accounts if acc.supports_notifications()]


	def _get_non_idle_accounts(self, accounts):
		return [acc for acc in self._accounts if not acc.supports_notifications()]


	def _load_plugins(self, cfg):
		enabled_lst = cfg.get('core', 'enabled_plugins').split(',')
		enabled_lst = [s for s in [s.strip() for s in enabled_lst] if s != '']
		self._plugins = Plugin.load_plugins(cfg, self, enabled_lst)
	
		for p in self._plugins:
			try:
				p.enable()
				logging.info("Successfully enabled plugin '%s'." % p.get_modname())
			except:
				logging.error("Failed to enable plugin '%s'." % p.get_modname())
	

	def _unload_plugins(self):
		if len(self._plugins) > 0:
			err = False
		
			for p in self._plugins:
				try:
					p.disable()
				except:
					err = True
					logging.error("Failed to disable plugin '%s'." % p.get_modname())
		
			if not err:
				logging.info('Plugins disabled successfully.')





# Copyright 2011 - 2020 Patrick Ulbrich <zulu99@gmx.net>
# Copyright 2020 Andreas Angerer
# Copyright 2011 Ralf Hersel <ralf.hersel@gmx.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#

import threading
import logging

from Mailnag.common.utils import try_call
from Mailnag.common.i18n import _
from Mailnag.common.plugins import HookTypes
from Mailnag.daemon.mails import MailSyncer


class MailChecker:
	def __init__(self, cfg, memorizer, hookreg, conntest, dbus_service):
		self._firstcheck = True # first check after startup
		self._mailcheck_lock = threading.Lock()
		self._mailsyncer = MailSyncer(cfg)
		self._mailbox_seen_flags = bool(cfg.get('core', 'mailbox_seen_flags'))
		self._memorizer = memorizer
		self._hookreg = hookreg
		self._conntest = conntest
		self._dbus_service = dbus_service
		self._count_on_last_check = 0
		
	
	def check(self, accounts):
		# make sure multiple threads (idler and polling thread) 
		# don't check for mails simultaneously.
		with self._mailcheck_lock:
			logging.info('Checking %s email account(s).' % len(accounts))
			
			for f in self._hookreg.get_hook_funcs(HookTypes.MAIL_CHECK):
				try_call( f )
				
			if self._conntest.is_offline():
				logging.warning('No internet connection.')
				return
			
			all_mails = self._mailsyncer.sync(accounts)
			unseen_mails = []
			new_mails = []
			seen_mails_by_account = {}
			
			for mail in all_mails:
				if self._memorizer.contains(mail.id): # mail was fetched before
					if self._memorizer.is_unseen(mail.id): # mail was not marked as seen
						unseen_mails.append(mail)
						if self._firstcheck:
							new_mails.append(mail)
					elif self._mailbox_seen_flags:
						# if the mail account supports tagging mails as seen (e.g. IMAP), 
						# mark the mail as seen on the server as well.
						if mail.account.supports_mark_as_seen():
							if not mail.account in seen_mails_by_account:
								seen_mails_by_account[mail.account] = []
							seen_mails_by_account[mail.account].append(mail)
				else: # mail is fetched the first time
					unseen_mails.append(mail)
					new_mails.append(mail)
			
			# Flag mails to seen on server
			for acc, mails in seen_mails_by_account.items():
				acc.mark_as_seen(mails)
			
			self._memorizer.sync(all_mails)
			self._memorizer.save()
			self._firstcheck = False
			
			# apply filter plugin hooks
			filtered_unseen_mails = unseen_mails
			for f in self._hookreg.get_hook_funcs(HookTypes.FILTER_MAILS):
				filtered_unseen_mails = try_call( lambda: f(filtered_unseen_mails), filtered_unseen_mails )
			
			filtered_new_mails = [m for m in new_mails if m in filtered_unseen_mails]
			
			if len(filtered_new_mails) > 0:
				self._dbus_service.signal_mails_added(filtered_new_mails, filtered_unseen_mails)
				
				for f in self._hookreg.get_hook_funcs(HookTypes.MAILS_ADDED):
					try_call( lambda: f(filtered_new_mails, filtered_unseen_mails) )
			elif len(filtered_unseen_mails) != self._count_on_last_check:
				self._dbus_service.signal_mails_removed(filtered_unseen_mails)
				
				for f in self._hookreg.get_hook_funcs(HookTypes.MAILS_REMOVED):
					try_call( lambda: f(filtered_unseen_mails) )
			
			self._count_on_last_check = len(filtered_unseen_mails)
		
		return



# Copyright 2011 - 2021 Patrick Ulbrich <zulu99@gmx.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#

import os
import xdg.BaseDirectory as bd
from configparser import RawConfigParser

mailnag_defaults = {
	'core':
	{
		'poll_interval'		 : '10',
		'imap_idle_timeout'	 : '10',
		'mailbox_seen_flags' : '1',
		'autostart'			 : '1',
		'connectivity_test'	 : 'networkmonitor',
		'enabled_plugins'	 : 'dbusplugin, soundplugin, libnotifyplugin'
	}
}

cfg_folder = os.path.join(bd.xdg_config_home, "mailnag")
cfg_file = os.path.join(cfg_folder, "mailnag.cfg")

def cfg_exists():
	return os.path.exists(cfg_file)


def read_cfg():
	cfg = RawConfigParser()
	cfg.read_dict(mailnag_defaults)

	if os.path.exists(cfg_file):
		cfg.read(cfg_file)

	return cfg


def write_cfg(cfg):
	if not os.path.exists(cfg_folder):
		os.makedirs(cfg_folder)
	
	with open(cfg_file, 'w') as configfile: cfg.write(configfile)



# Copyright 2011 - 2021 Patrick Ulbrich <zulu99@gmx.net>
# Copyright 2020 Andreas Angerer
# Copyright 2016 Timo Kankare <timo.kankare@iki.fi>
# Copyright 2016 Thomas Haider <t.haider@deprecate.de>
# Copyright 2011 Ralf Hersel <ralf.hersel@gmx.net>
# Copyright 2019 razer <razerraz@free.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#

"""Implementation for IMAP mailbox connection."""

import email
import logging
import re

from Mailnag.backends.base import MailboxBackend
import Mailnag.common.imaplib2 as imaplib
from Mailnag.common.imaplib2 import AUTH
from Mailnag.common.exceptions import InvalidOperationException
from Mailnag.common.mutf7 import encode_mutf7, decode_mutf7


class IMAPMailboxBackend(MailboxBackend):
	"""Implementation of IMAP mail boxes."""

	def __init__(self, name = '', user = '', password = '', oauth2string = '',
				 server = '', port = '', ssl = True, folders = [], idle=True, **kw):
		self.name = name
		self.user = user
		self.password = password
		self.oauth2string = oauth2string
		self.server = server
		self.port = port
		self.ssl = ssl # bool
		self.folders = [encode_mutf7(folder) for folder in folders]
		self.idle = idle
		self._conn = None


	def open(self):
		if self._conn != None:
			raise InvalidOperationException("Account is aready open")
		
		self._conn = self._connect()

		
	def close(self):
		# if conn has already been closed, don't try to close it again
		if self._conn != None:
			conn = self._conn
			self._conn = None
			self._disconnect(conn)


	def is_open(self):
		return self._conn != None


	def list_messages(self):
		self._ensure_open()
		
		conn = self._conn
		
		if len(self.folders) == 0:
			folder_list = [ 'INBOX' ]
		else:
			folder_list = self.folders

		for folder in folder_list:
			# select IMAP folder
			conn.select(f'"{folder}"', readonly = True)
			try:
				status, data = conn.uid('SEARCH', None, '(UNSEEN)') # ALL or UNSEEN
			except:
				logging.warning('Folder %s does not exist.', folder)
				continue

			if status != 'OK' or None in [d for d in data]:
				logging.debug('Folder %s in status %s | Data: %s', (folder, status, data))
				continue # Bugfix LP-735071
			for num in data[0].split():
				typ, msg_data = conn.uid('FETCH', num, '(BODY.PEEK[HEADER])') # header only (without setting READ flag)
				for response_part in msg_data:
					if isinstance(response_part, tuple):
						try:
							msg = email.message_from_bytes(response_part[1])
						except:
							logging.debug("Couldn't get IMAP message.")
							continue
						yield (folder, msg, { 'uid' : num.decode("utf-8"), 'folder' : folder })


	def request_folders(self):
		lst = []
		
		# Always create a new connection as an existing one may
		# be used for IMAP IDLE.
		conn = self._connect()

		try:
			status, data = conn.list()
		finally:
			self._disconnect(conn)
		
		for d in data:
			match = re.match(r'.+\s+("."|"?NIL"?)\s+"?([^"]+)"?$', d.decode('utf-8'))

			if match == None:
				logging.warning("Folder format not supported.")
			else:
				folder = match.group(2)
				lst.append(decode_mutf7(folder))
		
		return lst


	def supports_mark_as_seen(self):
		return True


	def mark_as_seen(self, mails):
		# Always create a new connection as an existing one may
		# be used for IMAP IDLE.
		conn = self._connect()
		
		try:
			sorted_mails = sorted(mails, key = lambda m : m.flags['folder'] if 'folder' in m.flags else '')
			last_folder = ''
			
			for m in sorted_mails:
				if ('uid' in m.flags) and ('folder' in m.flags):
					try:
						folder = m.flags['folder']
						if folder != last_folder:
							conn.select(f'"{folder}"', readonly = False)
							last_folder = folder
						status, data = conn.uid("STORE", m.flags['uid'], "+FLAGS", "(\Seen)")
					except:
						logging.warning("Failed to set mail with uid %s to seen on server (account: '%s').", m.flags['uid'], acc.name)

		finally:
			self._disconnect(conn)
	
	
	def supports_notifications(self):
		"""Returns True if mailbox supports notifications.
		IMAP mailbox supports notifications if idle parameter is True"""
		return self.idle


	def notify_next_change(self, callback=None, timeout=None):
		self._ensure_open()
		
		# register idle callback that is called whenever an idle event
		# arrives (new mail / mail deleted).
		# the callback is called after <idle_timeout> minutes at the latest.
		# gmail sends keepalive events every 5 minutes.

		# idle callback (runs on a further thread)
		def _idle_callback(args):
			# error = (error_type, error_val)
			response, callback_arg, error = args
			# call actual callback
			callback(error)
		
		self._conn.idle(callback = _idle_callback, timeout = timeout)


	def cancel_notifications(self):
		# NOTE: Don't throw if the connection is closed.
		# Analogous to close().
		# (Otherwise cleanup code like in Idler._idle() will fail)
		# self._ensure_open()
		
		try:
			if self._conn != None:
				# Exit possible active idle state.
				# (also calls idle_callback)
				self._conn.noop()
		except:
			pass
	
	
	def _connect(self):
		conn = None
		
		try:
			if self.ssl:
				if self.port == '':
					conn = imaplib.IMAP4_SSL(self.server)
				else:
					conn = imaplib.IMAP4_SSL(self.server, int(self.port))
			else:
				if self.port == '':
					conn = imaplib.IMAP4(self.server)
				else:
					conn = imaplib.IMAP4(self.server, int(self.port))
				
				if 'STARTTLS' in conn.capabilities:
					conn.starttls()
				else:
					logging.warning("Using unencrypted connection for account '%s'" % self.name)
				
			if self.oauth2string != '':
				conn.authenticate('XOAUTH2', lambda x: self.oauth2string)
			elif 'AUTH=CRAM-MD5' in conn.capabilities:
				# use CRAM-MD5 auth if available
				conn.login_cram_md5(self.user, self.password)
			else:
				conn.login(self.user, self.password)
		except:
			try:
				if conn != None:
					# conn.close() # allowed in SELECTED state only
					conn.logout()
			except:	pass
			raise # re-throw exception
		
		# notify_next_change() (IMAP IDLE) requires a selected folder
		if conn.state == AUTH:
			self._select_single_folder(conn)
		
		return conn


	def _disconnect(self, conn):
		try:
			conn.close()
		finally:		
			# Closing the connection may fail (e.g. wrong state), 
			# but resources need to be freed anyway.
			conn.logout()
	
	
	def _select_single_folder(self, conn):
		if len(self.folders) == 1:
			folder = self.folders[0]
		else:
			folder = "INBOX"
		conn.select(f'"{folder}"', readonly = True)
	
	def _ensure_open(self):
		if not self.is_open():
			raise InvalidOperationException("Account is not open")




# Copyright 2013 - 2020 Patrick Ulbrich <zulu99@gmx.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#

import dbus
import dbus.service
import logging
from Mailnag.common.dist_cfg import DBUS_BUS_NAME, DBUS_OBJ_PATH
from Mailnag.common.exceptions import InvalidOperationException

MAX_INT32 = ((0xFFFFFFFF // 2) - 1)


# DBUS server that exports Mailnag signals and methods 
class DBusService(dbus.service.Object):
	def __init__(self, mailnag_daemon):
		self._mails = []
		self._daemon = mailnag_daemon
		bus_name = dbus.service.BusName(DBUS_BUS_NAME, bus = dbus.SessionBus())
		dbus.service.Object.__init__(self, bus_name, DBUS_OBJ_PATH)
	

	@dbus.service.signal(dbus_interface = DBUS_BUS_NAME, signature = 'aa{sv}aa{sv}')
	def MailsAdded(self, new_mails, all_mails):
		pass
	
	
	@dbus.service.signal(dbus_interface = DBUS_BUS_NAME, signature = 'aa{sv}')
	def MailsRemoved(self, remaining_mails):
		pass
	
		
	@dbus.service.method(dbus_interface = DBUS_BUS_NAME, out_signature = 'aa{sv}')
	def GetMails(self):
		return self._mails
		
	
	@dbus.service.method(dbus_interface = DBUS_BUS_NAME, out_signature = 'u')
	def GetMailCount(self):
		return len(self._mails)


	@dbus.service.method(dbus_interface = DBUS_BUS_NAME)
	def Shutdown(self):
		try:
			self._daemon.shutdown()
		except InvalidOperationException:
			pass
	
	
	@dbus.service.method(dbus_interface = DBUS_BUS_NAME)
	def CheckForMails(self):
		try:
			self._daemon.check_for_mails()
		except InvalidOperationException:
			pass


	@dbus.service.method(dbus_interface = DBUS_BUS_NAME, in_signature = 's')
	def MarkMailAsRead(self, mail_id):
		self._mails = [m for m in self._mails if m['id'] != mail_id]
		try:
			self._daemon.mark_mail_as_read(mail_id)
		except InvalidOperationException:
			pass
			
	
	def signal_mails_added(self, new_mails, all_mails):
		conv_new_mails = self._convert_mails(new_mails)
		conv_all_mails = self._convert_mails(all_mails)
		self._mails = conv_all_mails
		self.MailsAdded(conv_new_mails, conv_all_mails)
		
		
	def signal_mails_removed(self, remaining_mails):
		conv_remaining_mails = self._convert_mails(remaining_mails)
		self._mails = conv_remaining_mails
		self.MailsRemoved(conv_remaining_mails)
	
		
	def _convert_mails(self, mails):
		converted_mails = []
		for m in mails:
			d = {}
			name, addr = m.sender
			
			if m.datetime > MAX_INT32:
				logging.warning('dbusservice: datetime out of range (mailnag dbus api uses int32 timestamps).')
				datetime = 0
			else:
				datetime = m.datetime
			
			d['datetime'] = datetime			# int32 (i)
			d['subject'] = m.subject			# string (s)
			d['sender_name'] = name				# string (s)
			d['sender_addr'] = addr				# string (s)
			d['account_name'] = m.account.name	# string (s)
			d['id'] = m.id						# string (s)
			
			converted_mails.append(d)
		
		return converted_mails
